@@ -2,7 +2,7 @@
  * License, v. 2.0. If a copy of the MPL was not distributed with this
  * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
 
-use compact_cow_str::CompactCowStr;
+use cow_rc_str::CowRcStr;
 use std::borrow::Cow;
 use tokenizer::Token;
 
@@ -31,7 +31,17 @@ macro_rules! size_of_test {
     }
 }
 
-// These assume 64-bit
+// Some of these assume 64-bit
 size_of_test!(token, Token, 32);
 size_of_test!(std_cow_str, Cow<'static, str>, 32);
-size_of_test!(compact_cow_str, CompactCowStr, 16);
+size_of_test!(cow_rc_str, CowRcStr, 16);
+
+size_of_test!(tokenizer, ::tokenizer::Tokenizer, 40);
+size_of_test!(parser_input, ::parser::ParserInput, 112);
+size_of_test!(parser, ::parser::Parser, 16);
+size_of_test!(source_position, ::SourcePosition, 8);
+size_of_test!(parser_state, ::ParserState, 24);
+
+size_of_test!(basic_parse_error, ::BasicParseError, 40);
+size_of_test!(parse_error_lower_bound, ::ParseError<()>, 48);
+size_of_test!(precise_parse_error_lower_bound, ::PreciseParseError<()>, 72);
