@@ -276,7 +276,17 @@ fn outer_block_end_consumed() {
 fn bad_url_slice_out_of_bounds() {
     let mut input = ParserInput::new("url(\u{1}\\");
     let mut parser = Parser::new(&mut input);
-    let _ = parser.next_including_whitespace_and_comments(); // This used to panic
+    let result = parser.next_including_whitespace_and_comments();  // This used to panic
+    assert_eq!(result, Ok(&Token::BadUrl("\u{1}\\".into())));
+}
+
+/// https://bugzilla.mozilla.org/show_bug.cgi?id=1383975
+#[test]
+fn bad_url_slice_not_at_char_boundary() {
+    let mut input = ParserInput::new("url(9\n۰");
+    let mut parser = Parser::new(&mut input);
+    let result = parser.next_including_whitespace_and_comments();  // This used to panic
+    assert_eq!(result, Ok(&Token::BadUrl("9\n۰".into())));
 }
 
 #[test]
@@ -457,28 +467,40 @@ fn serialize_rgba_two_digit_float_if_roundtrips() {
 
 #[test]
 fn line_numbers() {
-    let mut input = ParserInput::new("foo bar\nbaz\r\n\n\"a\\\r\nb\"");
+    let mut input = ParserInput::new(concat!(
+        "fo\\30\r\n",
+        "0o bar/*\n",
+        "*/baz\r\n",
+        "\n",
+        "url(\r\n",
+        "  u \r\n",
+        ")\"a\\\r\n",
+        "b\""
+    ));
     let mut input = Parser::new(&mut input);
     assert_eq!(input.current_source_location(), SourceLocation { line: 0, column: 0 });
-    assert_eq!(input.next_including_whitespace(), Ok(&Token::Ident("foo".into())));
-    assert_eq!(input.current_source_location(), SourceLocation { line: 0, column: 3 });
+    assert_eq!(input.next_including_whitespace(), Ok(&Token::Ident("fo00o".into())));
+    assert_eq!(input.current_source_location(), SourceLocation { line: 1, column: 2 });
     assert_eq!(input.next_including_whitespace(), Ok(&Token::WhiteSpace(" ")));
-    assert_eq!(input.current_source_location(), SourceLocation { line: 0, column: 4 });
-    assert_eq!(input.next_including_whitespace(), Ok(&Token::Ident("bar".into())));
-    assert_eq!(input.current_source_location(), SourceLocation { line: 0, column: 7 });
-    assert_eq!(input.next_including_whitespace(), Ok(&Token::WhiteSpace("\n")));
-    assert_eq!(input.current_source_location(), SourceLocation { line: 1, column: 0 });
-    assert_eq!(input.next_including_whitespace(), Ok(&Token::Ident("baz".into())));
     assert_eq!(input.current_source_location(), SourceLocation { line: 1, column: 3 });
-    let position = input.position();
+    assert_eq!(input.next_including_whitespace(), Ok(&Token::Ident("bar".into())));
+    assert_eq!(input.current_source_location(), SourceLocation { line: 1, column: 6 });
+    assert_eq!(input.next_including_whitespace_and_comments(), Ok(&Token::Comment("\n")));
+    assert_eq!(input.current_source_location(), SourceLocation { line: 2, column: 2 });
+    assert_eq!(input.next_including_whitespace(), Ok(&Token::Ident("baz".into())));
+    assert_eq!(input.current_source_location(), SourceLocation { line: 2, column: 5 });
+    let state = input.state();
 
     assert_eq!(input.next_including_whitespace(), Ok(&Token::WhiteSpace("\r\n\n")));
-    assert_eq!(input.current_source_location(), SourceLocation { line: 3, column: 0 });
+    assert_eq!(input.current_source_location(), SourceLocation { line: 4, column: 0 });
 
-    assert_eq!(input.source_location(position), SourceLocation { line: 1, column: 3 });
+    assert_eq!(state.source_location(), SourceLocation { line: 2, column: 5 });
+
+    assert_eq!(input.next_including_whitespace(), Ok(&Token::UnquotedUrl("u".into())));
+    assert_eq!(input.current_source_location(), SourceLocation { line: 6, column: 1 });
 
     assert_eq!(input.next_including_whitespace(), Ok(&Token::QuotedString("ab".into())));
-    assert_eq!(input.current_source_location(), SourceLocation { line: 4, column: 2 });
+    assert_eq!(input.current_source_location(), SourceLocation { line: 7, column: 2 });
     assert!(input.next_including_whitespace().is_err());
 }
 
@@ -691,21 +713,21 @@ impl<'i> DeclarationParser<'i> for JsonParser {
         let mut value = vec![];
         let mut important = false;
         loop {
-            let start_position = input.position();
+            let start = input.state();
             if let Ok(mut token) = input.next_including_whitespace().map(|t| t.clone()) {
                 // Hack to deal with css-parsing-tests assuming that
                 // `!important` in the middle of a declaration value is OK.
                 // This can never happen per spec
                 // (even CSS Variables forbid top-level `!`)
                 if token == Token::Delim('!') {
-                    input.reset(start_position);
+                    input.reset(&start);
                     if parse_important(input).is_ok() {
                         if input.is_exhausted() {
                             important = true;
                             break
                         }
                     }
-                    input.reset(start_position);
+                    input.reset(&start);
                     token = input.next_including_whitespace().unwrap().clone();
                 }
                 value.push(one_component_value_to_json(token, input));
