@@ -5,11 +5,11 @@
 // https://drafts.csswg.org/css-syntax/#tokenization
 
 use std::ops::Range;
-use std::cell::Cell;
 use std::char;
 use std::ascii::AsciiExt;
 use std::i32;
 
+use parser::ParserState;
 use cow_rc_str::CowRcStr;
 use self::Token::*;
 
@@ -205,8 +205,8 @@ pub struct Tokenizer<'a> {
     input: &'a str,
     /// Counted in bytes, not code points. From 0.
     position: usize,
-    /// Cache for `source_location()`
-    last_known_source_location: Cell<(SourcePosition, SourceLocation)>,
+    current_line_start_position: usize,
+    current_line_number: u32,
     var_functions: SeenStatus,
     viewport_percentages: SeenStatus,
 }
@@ -222,11 +222,16 @@ enum SeenStatus {
 impl<'a> Tokenizer<'a> {
     #[inline]
     pub fn new(input: &str) -> Tokenizer {
+        Tokenizer::with_first_line_number(input, 0)
+    }
+
+    #[inline]
+    pub fn with_first_line_number(input: &str, first_line_number: u32) -> Tokenizer {
         Tokenizer {
             input: input,
             position: 0,
-            last_known_source_location: Cell::new((SourcePosition(0),
-                                                   SourceLocation { line: 0, column: 0 })),
+            current_line_start_position: 0,
+            current_line_number: first_line_number,
             var_functions: SeenStatus::DontCare,
             viewport_percentages: SeenStatus::DontCare,
         }
@@ -288,8 +293,28 @@ impl<'a> Tokenizer<'a> {
     }
 
     #[inline]
-    pub fn reset(&mut self, new_position: SourcePosition) {
-        self.position = new_position.0;
+    pub fn current_source_location(&self) -> SourceLocation {
+        SourceLocation {
+            line: self.current_line_number,
+            column: (self.position - self.current_line_start_position) as u32,
+        }
+    }
+
+    #[inline]
+    pub fn state(&self) -> ParserState {
+        ParserState {
+            position: self.position,
+            current_line_start_position: self.current_line_start_position,
+            current_line_number: self.current_line_number,
+            at_start_of: None,
+        }
+    }
+
+    #[inline]
+    pub fn reset(&mut self, state: &ParserState) {
+        self.position = state.position;
+        self.current_line_start_position = state.current_line_start_position;
+        self.current_line_number = state.current_line_number;
     }
 
     #[inline]
@@ -302,12 +327,6 @@ impl<'a> Tokenizer<'a> {
         &self.input[range.start.0..range.end.0]
     }
 
-    #[inline]
-    pub fn current_source_location(&self) -> SourceLocation {
-        let position = SourcePosition(self.position);
-        self.source_location(position)
-    }
-
     pub fn current_source_line(&self) -> &'a str {
         let current = self.position;
         let start = self.input[0..current]
@@ -319,37 +338,6 @@ impl<'a> Tokenizer<'a> {
         &self.input[start..end]
     }
 
-    pub fn source_location(&self, position: SourcePosition) -> SourceLocation {
-        let target = position.0;
-        let mut location;
-        let mut position;
-        let (SourcePosition(last_known_position), last_known_location) =
-            self.last_known_source_location.get();
-        if target >= last_known_position {
-            position = last_known_position;
-            location = last_known_location;
-        } else {
-            // For now we’re only traversing the source *forwards* to count newlines.
-            // So if the requested position is before the last known one,
-            // start over from the beginning.
-            position = 0;
-            location = SourceLocation { line: 0, column: 0 };
-        }
-        let mut source = &self.input[position..target];
-        while let Some(newline_position) = source.find(|c| matches!(c, '\n' | '\r' | '\x0C')) {
-            let offset = newline_position +
-                if source[newline_position..].starts_with("\r\n") { 2 } else { 1 };
-            source = &source[offset..];
-            position += offset;
-            location.line += 1;
-            location.column = 0;
-        }
-        debug_assert!(position <= target);
-        location.column += (target - position) as u32;
-        self.last_known_source_location.set((SourcePosition(target), location));
-        location
-    }
-
     #[inline]
     pub fn next_byte(&self) -> Option<u8> {
         if self.is_eof() {
@@ -391,6 +379,14 @@ impl<'a> Tokenizer<'a> {
         self.input[self.position..].chars().next().unwrap()
     }
 
+    fn seen_newline(&mut self, is_cr: bool) {
+        if is_cr && self.next_byte() == Some(/* LF */ b'\n') {
+            return
+        }
+        self.current_line_start_position = self.position;
+        self.current_line_number += 1;
+    }
+
     #[inline]
     fn has_newline_at(&self, offset: usize) -> bool {
         self.position + offset < self.input.len() &&
@@ -410,15 +406,15 @@ impl<'a> Tokenizer<'a> {
     }
 }
 
-
+/// A position from the start of the input, counted in UTF-8 bytes.
 #[derive(PartialEq, Eq, PartialOrd, Ord, Debug, Clone, Copy)]
-pub struct SourcePosition(usize);
+pub struct SourcePosition(pub(crate) usize);
 
 
 /// The line and column number for a given position within the input.
 #[derive(PartialEq, Eq, Debug, Clone, Copy)]
 pub struct SourceLocation {
-    /// The line number, starting at 0 for the first line.
+    /// The line number, starting at 0 for the first line, unless `with_first_line_number` was used.
     pub line: u32,
 
     /// The column number within a line, starting at 0 for first the character of the line.
@@ -432,16 +428,14 @@ fn next_token<'a>(tokenizer: &mut Tokenizer<'a>) -> Result<Token<'a>, ()> {
     }
     let b = tokenizer.next_byte_unchecked();
     let token = match_byte! { b,
-        b'\t' | b'\n' | b' ' | b'\r' | b'\x0C' => {
-            let start_position = tokenizer.position();
-            tokenizer.advance(1);
-            while !tokenizer.is_eof() {
-                match tokenizer.next_byte_unchecked() {
-                    b' ' | b'\t' | b'\n' | b'\r' | b'\x0C' => tokenizer.advance(1),
-                    _ => break,
-                }
-            }
-            WhiteSpace(tokenizer.slice_from(start_position))
+        b' ' | b'\t' => {
+            consume_whitespace(tokenizer, false, false)
+        },
+        b'\n' | b'\x0C' => {
+            consume_whitespace(tokenizer, true, false)
+        },
+        b'\r' => {
+            consume_whitespace(tokenizer, true, true)
         },
         b'"' => { consume_string(tokenizer, false) },
         b'#' => {
@@ -513,21 +507,7 @@ fn next_token<'a>(tokenizer: &mut Tokenizer<'a>) -> Result<Token<'a>, ()> {
         }
         b'/' => {
             if tokenizer.starts_with(b"/*") {
-                tokenizer.advance(2);  // consume "/*"
-                let start_position = tokenizer.position();
-                let content;
-                match tokenizer.input[tokenizer.position..].find("*/") {
-                    Some(offset) => {
-                        tokenizer.advance(offset);
-                        content = tokenizer.slice_from(start_position);
-                        tokenizer.advance(2);
-                    }
-                    None => {
-                        tokenizer.position = tokenizer.input.len();
-                        content = tokenizer.slice_from(start_position);
-                    }
-                }
-                Comment(content)
+                Comment(consume_comment(tokenizer))
             } else {
                 tokenizer.advance(1);
                 Delim('/')
@@ -585,6 +565,64 @@ fn next_token<'a>(tokenizer: &mut Tokenizer<'a>) -> Result<Token<'a>, ()> {
 }
 
 
+fn consume_whitespace<'a>(tokenizer: &mut Tokenizer<'a>, newline: bool, is_cr: bool) -> Token<'a> {
+    let start_position = tokenizer.position();
+    tokenizer.advance(1);
+    if newline {
+        tokenizer.seen_newline(is_cr)
+    }
+    while !tokenizer.is_eof() {
+        let b = tokenizer.next_byte_unchecked();
+        match_byte! { b,
+            b' ' | b'\t' => {
+                tokenizer.advance(1);
+            }
+            b'\n' | b'\x0C' => {
+                tokenizer.advance(1);
+                tokenizer.seen_newline(false);
+            }
+            b'\r' => {
+                tokenizer.advance(1);
+                tokenizer.seen_newline(true);
+            }
+            _ => {
+                break
+            }
+        }
+    }
+    WhiteSpace(tokenizer.slice_from(start_position))
+}
+
+
+fn consume_comment<'a>(tokenizer: &mut Tokenizer<'a>) -> &'a str {
+    tokenizer.advance(2);  // consume "/*"
+    let start_position = tokenizer.position();
+    while !tokenizer.is_eof() {
+        match_byte! { tokenizer.next_byte_unchecked(),
+            b'*' => {
+                let end_position = tokenizer.position();
+                tokenizer.advance(1);
+                if tokenizer.next_byte() == Some(b'/') {
+                    tokenizer.advance(1);
+                    return tokenizer.slice(start_position..end_position)
+                }
+            }
+            b'\n' | b'\x0C' => {
+                tokenizer.advance(1);
+                tokenizer.seen_newline(false);
+            }
+            b'\r' => {
+                tokenizer.advance(1);
+                tokenizer.seen_newline(true);
+            }
+            _ => {
+                tokenizer.advance(1);
+            }
+        }
+    }
+    tokenizer.slice_from(start_position)
+}
+
 fn consume_string<'a>(tokenizer: &mut Tokenizer<'a>, single_quote: bool) -> Token<'a> {
     match consume_quoted_string(tokenizer, single_quote) {
         Ok(value) => QuotedString(value),
@@ -661,12 +699,19 @@ fn consume_quoted_string<'a>(tokenizer: &mut Tokenizer<'a>, single_quote: bool)
                 if !tokenizer.is_eof() {
                     match tokenizer.next_byte_unchecked() {
                         // Escaped newline
-                        b'\n' | b'\x0C' => tokenizer.advance(1),
+                        b'\n' | b'\x0C' => {
+                            tokenizer.advance(1);
+                            tokenizer.seen_newline(false);
+                        }
                         b'\r' => {
                             tokenizer.advance(1);
                             if tokenizer.next_byte() == Some(b'\n') {
                                 tokenizer.advance(1);
                             }
+                            // `is_cr = true` is useful to skip \r when the next iteration
+                            // of a loop will call `seen_newline` again for the following \n.
+                            // In this case we’re consuming both in this iteration, so passing `false`.
+                            tokenizer.seen_newline(false);
                         }
                         // This pushes one well-formed code point
                         _ => consume_escape_and_write(tokenizer, &mut string_bytes)
@@ -933,24 +978,57 @@ unsafe fn from_utf8_release_unchecked(string_bytes: Vec<u8>) -> String {
 
 fn consume_unquoted_url<'a>(tokenizer: &mut Tokenizer<'a>) -> Result<Token<'a>, ()> {
     // This is only called after "url(", so the current position is a code point boundary.
-    for (offset, c) in tokenizer.input[tokenizer.position..].bytes().enumerate() {
-        match_byte! { c,
-            b' ' | b'\t' | b'\n' | b'\r' | b'\x0C' => {},
+    let start_position = tokenizer.position;
+    let from_start = &tokenizer.input[tokenizer.position..];
+    let mut newlines = 0;
+    let mut last_newline = 0;
+    let mut found_printable_char = false;
+    let mut iter = from_start.bytes().enumerate();
+    loop {
+        let (offset, b) = match iter.next() {
+            Some(item) => item,
+            None => {
+                tokenizer.position = tokenizer.input.len();
+                break
+            }
+        };
+        match_byte! { b,
+            b' ' | b'\t' => {},
+            b'\n' | b'\x0C' => {
+                newlines += 1;
+                last_newline = offset;
+            }
+            b'\r' => {
+                if from_start.as_bytes().get(offset + 1) != Some(&b'\n') {
+                    newlines += 1;
+                    last_newline = offset;
+                }
+            }
             b'"' | b'\'' => { return Err(()) },  // Do not advance
             b')' => {
                 tokenizer.advance(offset + 1);
-                return Ok(UnquotedUrl("".into()));
+                break
             }
             _ => {
                 tokenizer.advance(offset);
-                // This function only consumed ASCII (whitespace) bytes,
-                // so the current position is a code point boundary.
-                return Ok(consume_unquoted_url_internal(tokenizer))
+                found_printable_char = true;
+                break
             }
         }
     }
-    tokenizer.position = tokenizer.input.len();
-    return Ok(UnquotedUrl("".into()));
+
+    if newlines > 0 {
+        tokenizer.current_line_number += newlines;
+        tokenizer.current_line_start_position = start_position + last_newline + 1;
+    }
+
+    if found_printable_char {
+        // This function only consumed ASCII (whitespace) bytes,
+        // so the current position is a code point boundary.
+        return Ok(consume_unquoted_url_internal(tokenizer))
+    } else {
+        return Ok(UnquotedUrl("".into()))
+    }
 
     fn consume_unquoted_url_internal<'a>(tokenizer: &mut Tokenizer<'a>) -> Token<'a> {
         // This function is only called with start_pos at a code point boundary.
@@ -963,8 +1041,7 @@ fn consume_unquoted_url<'a>(tokenizer: &mut Tokenizer<'a>) -> Result<Token<'a>,
             match_byte! { tokenizer.next_byte_unchecked(),
                 b' ' | b'\t' | b'\n' | b'\r' | b'\x0C' => {
                     let value = tokenizer.slice_from(start_pos);
-                    tokenizer.advance(1);
-                    return consume_url_end(tokenizer, value.into())
+                    return consume_url_end(tokenizer, start_pos, value.into())
                 }
                 b')' => {
                     let value = tokenizer.slice_from(start_pos);
@@ -974,7 +1051,7 @@ fn consume_unquoted_url<'a>(tokenizer: &mut Tokenizer<'a>) -> Result<Token<'a>,
                 b'\x01'...b'\x08' | b'\x0B' | b'\x0E'...b'\x1F' | b'\x7F'  // non-printable
                     | b'"' | b'\'' | b'(' => {
                     tokenizer.advance(1);
-                    return consume_bad_url(tokenizer)
+                    return consume_bad_url(tokenizer, start_pos)
                 },
                 b'\\' | b'\0' => {
                     // * The tokenizer’s input is UTF-8 since it’s `&str`.
@@ -986,29 +1063,28 @@ fn consume_unquoted_url<'a>(tokenizer: &mut Tokenizer<'a>) -> Result<Token<'a>,
                     break
                 }
                 _ => {
-                    tokenizer.consume_byte();
+                    tokenizer.advance(1);
                 }
             }
         }
         while !tokenizer.is_eof() {
             match_byte! { tokenizer.consume_byte(),
                 b' ' | b'\t' | b'\n' | b'\r' | b'\x0C' => {
-                    return consume_url_end(
-                        tokenizer,
-                        // string_bytes is well-formed UTF-8, see other comments.
-                        unsafe { from_utf8_release_unchecked(string_bytes) }.into()
-                    )
+                    // string_bytes is well-formed UTF-8, see other comments.
+                    let string = unsafe { from_utf8_release_unchecked(string_bytes) }.into();
+                    tokenizer.position -= 1;
+                    return consume_url_end(tokenizer, start_pos, string)
                 }
                 b')' => {
                     break;
                 }
                 b'\x01'...b'\x08' | b'\x0B' | b'\x0E'...b'\x1F' | b'\x7F'  // non-printable
                     | b'"' | b'\'' | b'(' => {
-                    return consume_bad_url(tokenizer);
+                    return consume_bad_url(tokenizer, start_pos);
                 }
                 b'\\' => {
                     if tokenizer.has_newline_at(0) {
-                        return consume_bad_url(tokenizer)
+                        return consume_bad_url(tokenizer, start_pos)
                     }
 
                     // This pushes one well-formed code point to string_bytes
@@ -1028,30 +1104,48 @@ fn consume_unquoted_url<'a>(tokenizer: &mut Tokenizer<'a>) -> Result<Token<'a>,
         )
     }
 
-    fn consume_url_end<'a>(tokenizer: &mut Tokenizer<'a>, string: CowRcStr<'a>) -> Token<'a> {
+    fn consume_url_end<'a>(tokenizer: &mut Tokenizer<'a>,
+                           start_pos: SourcePosition,
+                           string: CowRcStr<'a>)
+                           -> Token<'a> {
         while !tokenizer.is_eof() {
             match_byte! { tokenizer.consume_byte(),
-                b' ' | b'\t' | b'\n' | b'\r' | b'\x0C' => {},
-                b')' => { break },
+                b')' => {
+                    break
+                }
+                b' ' | b'\t' => {}
+                b'\n' | b'\x0C' => {
+                    tokenizer.seen_newline(false);
+                }
+                b'\r' => {
+                    tokenizer.seen_newline(true);
+                }
                 _ => {
-                    return consume_bad_url(tokenizer);
+                    return consume_bad_url(tokenizer, start_pos);
                 }
             }
         }
         UnquotedUrl(string)
     }
 
-    fn consume_bad_url<'a>(tokenizer: &mut Tokenizer<'a>) -> Token<'a> {
-        let start_pos = tokenizer.position();
+    fn consume_bad_url<'a>(tokenizer: &mut Tokenizer<'a>, start_pos: SourcePosition) -> Token<'a> {
         // Consume up to the closing )
         while !tokenizer.is_eof() {
             match_byte! { tokenizer.consume_byte(),
-                b')' => { break },
+                b')' => {
+                    break
+                }
                 b'\\' => {
                     if matches!(tokenizer.next_byte(), Some(b')') | Some(b'\\')) {
                         tokenizer.advance(1); // Skip an escaped ')' or '\'
                     }
                 }
+                b'\n' | b'\x0C' => {
+                    tokenizer.seen_newline(false);
+                }
+                b'\r' => {
+                    tokenizer.seen_newline(true);
+                }
                 _ => {},
             }
         }
@@ -1092,15 +1186,22 @@ fn consume_escape(tokenizer: &mut Tokenizer) -> char {
         b'0'...b'9' | b'A'...b'F' | b'a'...b'f' => {
             let (c, _) = consume_hex_digits(tokenizer);
             if !tokenizer.is_eof() {
-                match tokenizer.next_byte_unchecked() {
-                    b' ' | b'\t' | b'\n' | b'\x0C' => tokenizer.advance(1),
+                match_byte! { tokenizer.next_byte_unchecked(),
+                    b' ' | b'\t' => {
+                        tokenizer.advance(1)
+                    }
+                    b'\n' | b'\x0C' => {
+                        tokenizer.advance(1);
+                        tokenizer.seen_newline(false)
+                    }
                     b'\r' => {
                         tokenizer.advance(1);
                         if !tokenizer.is_eof() && tokenizer.next_byte_unchecked() == b'\n' {
                             tokenizer.advance(1);
                         }
+                        tokenizer.seen_newline(false)
                     }
-                    _ => ()
+                    _ => {}
                 }
             }
             static REPLACEMENT_CHAR: char = '\u{FFFD}';
